/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.39 from the
 * contents of Fanotify.xs. Do not edit this file, edit Fanotify.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "Fanotify.xs"
/*
 * Copyright (C) 2014 Bastian Friedrich
 * <bastian.friedrich@collax.com> / <bastian@friedrich.link>
 */
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include <pwd.h>
#include <unistd.h>
#include <sys/fanotify.h>

#include "const-c.inc"

/*
 *******************************************************************************
 * Code documentation for perl methods and functions can be found in the
 * module's pod
 *******************************************************************************
 */

#define PERLCLASS_NOTGRP "Linux::Fanotify::FanotifyGroup"
#define PERLCLASS_EVENT "Linux::Fanotify::Event"

/*
 * We don't want to expose functionality to write one event's fd to a
 * notification group where it did not originate. fanotify_bundle
 * is the event metadata, plus its originating notification group
 */

struct fanotify_bundle {
	struct fanotify_event_metadata metadata;
	int fd;	/* Notification group fd */
	unsigned short int needs_response;
};

/*
 * Convert a perl SV for a notification group to a file descriptor
 */
int
notgrp2fd(SV *notgrp, int *dst) {
	SV *sv;

	if (!SvROK(notgrp)) {
		/* No RV */
		return 0;
	}
	if (!sv_isobject(notgrp) || !sv_isa(notgrp, PERLCLASS_NOTGRP)) {
		/* No notification group obj */
		return 0;
	}

	sv = SvRV(notgrp);

	if (!SvIOK(sv)) {
		/* No int */
		return 0;
	}

	*dst = SvIV(sv);

	return 1;
}

/*
 * Convert a perl SV for a fanotify_bundle (metadata plus fanotify group fd) to
 * a fanotify_bundle structure
 */
struct fanotify_bundle *
event2bundle(SV *event) {
	SV *deref;

	if (!SvROK(event)) {
		/* No RV */
		return NULL;
	}
	if (!sv_isobject(event) || !sv_isa(event, PERLCLASS_EVENT)) {
		/* No notification group obj */
		return NULL;
	}

	deref = SvRV(event);

	if (!deref) {
		return NULL;
	}

	return (struct fanotify_bundle *)SvPV_nolen(deref);

}

/*
 * Convert a notification event bundle to a string.
 * Intended for debugging purposes only.
 */
char *
event2str(const struct fanotify_bundle *bundle, char *buf, int len) {

	const struct fanotify_event_metadata *event = &(bundle->metadata);

	char *fmt = "event at %p is: event_len = %d, vers = %d, reserved = %d, meta_len = %d, mask = %lld, fd = %d, pid = %d (fanotify group fd %d)";

	if (snprintf(buf, len, fmt,
		event, event->event_len, event->vers, event->reserved, event->metadata_len, event->mask, event->fd, event->pid, bundle->fd) >= len) {
		buf[len-1] = 0;
	}

	return buf;
}

/*
 * Print an event's contents on the screen
 */
void
dumpEvent(const struct fanotify_bundle *bundle) {
	char buf[256];

	event2str(bundle, buf, sizeof(buf));
	printf("%s\n", buf);
}

int
_event_write_response(SV *event, int response) {
	struct fanotify_bundle *bundle;
	struct fanotify_response rsp;
	int fd;
	int ret;

	if (!(bundle = event2bundle(event))) {
		Perl_croak(aTHX_ "Invalid event object");
	}

	if (!bundle->needs_response) {
		Perl_croak(aTHX_ "Event already responded to or non-respondable event");
	}

	/* only ALLOW and DENY are available/sensible responses */
	if ((response != FAN_ALLOW) && (response != FAN_DENY)) {
		Perl_croak(aTHX_ "Response is neither FAN_ALLOW nor FAN_DENY; refusing to write invalid response");
	}

	fd = bundle->fd;

	rsp.fd = bundle->metadata.fd;
	rsp.response = response;

	ret = write(fd, &rsp, sizeof(struct fanotify_response));

	if (ret >= 0) {
		bundle->needs_response = 0;
	}

	return ret;
}

int
_event_close(SV *self) {
	struct fanotify_bundle *bundle;
	int ret;
	SV *sv_default_response;
	int default_response = FAN_DENY;

	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}

	if ((sv_default_response = get_sv("Linux::Fanotify::default_response", 0))) {
		default_response = SvIV(sv_default_response);
		if (default_response == -1) {
			default_response = FAN_DENY;
		}
	}

	if (bundle->needs_response) {
		if (default_response) {
			_event_write_response(self, default_response);
		}
	}

	if (bundle->metadata.fd > 0) {
		ret = close(bundle->metadata.fd);
		if (ret == 0) {
			bundle->metadata.fd = -1;
		}
	} else {
		errno = EBADF;
		ret = -1;
	}
	return ret;
}

/*
 *******************************************************************************
 * Main package: Linux::Fanotify
 */
#line 205 "Fanotify.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

#line 349 "Fanotify.c"

/* INCLUDE:  Including 'const-xs.inc' from 'Fanotify.xs' */


XS_EUPXS(XS_Linux__Fanotify_AUTOLOAD); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify_AUTOLOAD)
{
    dVAR; dXSARGS;
    if (items != 0)
       croak_xs_usage(cv,  "");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
#line 185 "./const-xs.inc"
	SV *sv = newSVpvn_flags(SvPVX(cv), SvCUR(cv), SVs_TEMP | SvUTF8(cv));
	const COP *cop = PL_curcop;
#line 366 "Fanotify.c"
#line 188 "./const-xs.inc"
#ifndef SYMBIAN
	/* It's not obvious how to calculate this at C pre-processor time.
	   However, any compiler optimiser worth its salt should be able to
	   remove the dead code, and hopefully the now-obviously-unused static
	   function too.  */
	HV *constant_missing = (C_ARRAY_LENGTH(values_for_notfound) > 1)
	    ? get_missing_hash(aTHX) : NULL;
	if ((C_ARRAY_LENGTH(values_for_notfound) > 1)
	    ? hv_exists_ent(constant_missing, sv, 0) : 0) {
	    sv = newSVpvf("Your vendor has not defined Linux::Fanotify macro %" SVf
			  ", used at %" COP_FILE_F " line %" UVuf "\n", 
			  sv, COP_FILE(cop), (UV)CopLINE(cop));
	} else
#endif
	{
	    sv = newSVpvf("%" SVf
                          " is not a valid Linux::Fanotify macro at %"
			  COP_FILE_F " line %" UVuf "\n",
			  sv, COP_FILE(cop), (UV)CopLINE(cop));
	}
	croak_sv(sv_2mortal(sv));
#line 389 "Fanotify.c"
	PUTBACK;
	return;
    }
}


/* INCLUDE: Returning to 'Fanotify.xs' from 'const-xs.inc' */


XS_EUPXS(XS_Linux__Fanotify_fanotify_init); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify_fanotify_init)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "flags, event_f_flags");
    {
	unsigned int	flags = (unsigned int)SvUV(ST(0))
;
	unsigned int	event_f_flags = (unsigned int)SvUV(ST(1))
;
	SV *	RETVAL;
#line 208 "Fanotify.xs"
	int fd;
	SV *notgrp;
#line 414 "Fanotify.c"
#line 211 "Fanotify.xs"
	fd = fanotify_init(flags, event_f_flags);

	if (fd == -1) {
		XSRETURN_UNDEF;
	}

	notgrp = newSV(0);

	sv_setref_iv(notgrp, PERLCLASS_NOTGRP, fd);

	SvREADONLY_on(SvRV(notgrp));

	RETVAL = notgrp;
#line 429 "Fanotify.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify_fanotify_mark); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify_fanotify_mark)
{
    dVAR; dXSARGS;
    if (items < 2 || items > 5)
       croak_xs_usage(cv,  "notgrp, flags, mask = 0, dirfd = 0, pathname = NULL");
    {
	SV *	notgrp = ST(0)
;
	unsigned int	flags = (unsigned int)SvUV(ST(1))
;
	uint64_t	mask;
	int	dirfd;
	const char *	pathname;
	int	RETVAL;
	dXSTARG;

	if (items < 3)
	    mask = 0;
	else {
	    mask = (uint64_t)SvUV(ST(2))
;
	}

	if (items < 4)
	    dirfd = 0;
	else {
	    dirfd = (int)SvIV(ST(3))
;
	}

	if (items < 5)
	    pathname = NULL;
	else {
	    pathname = (const char *)SvPV_nolen(ST(4))
;
	}
#line 238 "Fanotify.xs"
	int ret;
	int fd = -1;
#line 477 "Fanotify.c"
#line 241 "Fanotify.xs"
	if (!notgrp2fd(notgrp, &fd)) {
		Perl_croak(aTHX_ "Invalid fanotify_fd");
	} else {
		ret = fanotify_mark(fd, flags, mask, dirfd, pathname);

		if (ret == -1) {
			XSRETURN_UNDEF;
		}
	}
	RETVAL = 1;
#line 489 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify_fanotify_read); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify_fanotify_read)
{
    dVAR; dXSARGS;
    if (items < 1 || items > 2)
       croak_xs_usage(cv,  "notgrp, max = 0");
    PERL_UNUSED_VAR(ax); /* -Wall */
    SP -= items;
    {
	SV *	notgrp = ST(0)
;
	int	max;

	if (items < 2)
	    max = 0;
	else {
	    max = (int)SvIV(ST(1))
;
	}
#line 262 "Fanotify.xs"
	struct fanotify_event_metadata *buf;
	const struct fanotify_event_metadata *metadata;
	struct fanotify_bundle *bundle;
	ssize_t len;
	int size;
	int fd;
	SV *event;
	SV *eventref;
#line 524 "Fanotify.c"
#line 271 "Fanotify.xs"
	/*
	 * Check arguments and allocate read buffer
	 */
	if (max < 1) {
		max = 170;	// Results in a buffer slightly smaller than 4k
	}
	if (max > 4096) {	// 4096 results in an almost 100k buffer. Don't allow more.
		Perl_croak(aTHX_ "Maximum buffer size exceeded (max = 4096)");
	}
	if (!notgrp2fd(notgrp, &fd)) {
		Perl_croak(aTHX_ "Invalid notification group object");
	}

	size = max * sizeof(struct fanotify_event_metadata);
	buf = (struct fanotify_event_metadata *)malloc(size + (sizeof(struct fanotify_bundle) - sizeof(struct fanotify_event_metadata)));

	if (!buf) {
		Perl_croak(aTHX_ "Could not allocate memory");
	}

	/*
	 * Read from fanotify queue
	 */


	len = read(fd, (void *)buf, size);

	if (len == -1) {
		/* Read error. errno is already set, we simply return an empty list and expect the user to check errno */
		free(buf);
		XSRETURN_EMPTY;
	}

	// printf("Number of event structures: %d\n",
	// 	len/sizeof(struct fanotify_event_metadata));

	metadata = buf;

	while (FAN_EVENT_OK(metadata, len)) {
		if (metadata->vers != FANOTIFY_METADATA_VERSION) {
			Perl_croak(aTHX_ "Mismatch of fanotify metadata version.");
		}

		event = newSVpvn((const char *)metadata, sizeof(struct fanotify_bundle));
		sv_2mortal(event);

		bundle = (struct fanotify_bundle *)SvPV_nolen(event);
		bundle->fd = fd;
		if (metadata->mask & (FAN_ACCESS_PERM | FAN_OPEN_PERM)) {
			bundle->needs_response = 1;
		} else {
			bundle->needs_response = 0;
		}

		eventref = newRV_noinc(event);
		sv_bless(eventref, gv_stashpv(PERLCLASS_EVENT, GV_ADD | SVf_UTF8));
		SvREADONLY_on(eventref);

		XPUSHs(eventref);

		metadata = FAN_EVENT_NEXT(metadata, len);
	}

	free(buf);
#line 590 "Fanotify.c"
	PUTBACK;
	return;
    }
}


XS_EUPXS(XS_Linux__Fanotify_fanotify_write); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify_fanotify_write)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "event, response");
    {
	SV *	event = ST(0)
;
	int	response = (int)SvIV(ST(1))
;
	int	RETVAL;
	dXSTARG;
#line 344 "Fanotify.xs"
	int ret;
#line 612 "Fanotify.c"
#line 346 "Fanotify.xs"
	ret = _event_write_response(event, response);

	if (ret == -1) {
		XSRETURN_UNDEF;
	}

	RETVAL = ret;

#line 622 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__FanotifyGroup_getfd); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__FanotifyGroup_getfd)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	int	RETVAL;
	dXSTARG;
#line 370 "Fanotify.xs"
	int fd;
#line 642 "Fanotify.c"
#line 372 "Fanotify.xs"
	if (!notgrp2fd(self, &fd)) {
		Perl_croak(aTHX_ "Invalid notification group object");
	}

	if (fd > 0) {
		RETVAL = fd;
	} else {
		XSRETURN_UNDEF;
	}
#line 653 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__FanotifyGroup_close); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__FanotifyGroup_close)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	int	RETVAL;
	dXSTARG;
#line 391 "Fanotify.xs"
	int fd;
	SV *deref;
	int ret;
#line 675 "Fanotify.c"
#line 395 "Fanotify.xs"
	if (!notgrp2fd(self, &fd)) {
		Perl_croak(aTHX_ "Invalid notification group object");
	}

	if (fd > 0) {
		ret = close(fd);

		deref = SvRV(self);
		SvIV_set(deref, -1);
	} else {
		ret = -1;
		errno = EBADF;
	}

	// Convert C style 0/-1 to perl style 1/undef
	if (ret == -1) {
		XSRETURN_UNDEF;
	} else {
		ret = 1;
	}
	RETVAL = ret;
#line 698 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__FanotifyGroup_DESTROY); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__FanotifyGroup_DESTROY)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
#line 426 "Fanotify.xs"
	int fd;
	int autoclose = 1;
	SV *sv_autoclose;
#line 718 "Fanotify.c"
#line 430 "Fanotify.xs"
	if (!notgrp2fd(self, &fd)) {
		Perl_croak(aTHX_ "Invalid notification group object");
	}

	if (fd > 0) {
		if ((sv_autoclose = get_sv("Linux::Fanotify::FanotifyGroup::autoclose", 0))) {
			autoclose = SvIV(sv_autoclose);
		}
		if (autoclose) {
			close(fd);
		}
	}
#line 732 "Fanotify.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Linux__Fanotify__Event_close); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_close)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	int	RETVAL;
	dXSTARG;
#line 456 "Fanotify.xs"
	int ret;
#line 751 "Fanotify.c"
#line 458 "Fanotify.xs"
	ret = _event_close(self);

	if (ret == -1) {
		XSRETURN_UNDEF;
	} else {
		ret = 1;
	}

	RETVAL = ret;
#line 762 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_DESTROY); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_DESTROY)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
#line 477 "Fanotify.xs"
	int autoclose = 1;
	SV *sv_autoclose;
#line 781 "Fanotify.c"
#line 480 "Fanotify.xs"
	if ((sv_autoclose = get_sv("Linux::Fanotify::Event::autoclose", 0))) {
		autoclose = SvIV(sv_autoclose);
	}
	if (autoclose) {
		_event_close(self);
	}
#line 789 "Fanotify.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Linux__Fanotify__Event_needsResponse); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_needsResponse)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	int	RETVAL;
	dXSTARG;
#line 494 "Fanotify.xs"
	struct fanotify_bundle *bundle;
#line 808 "Fanotify.c"
#line 496 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}

	RETVAL = bundle->needs_response;
#line 815 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event__write); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event__write)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "self, response");
    {
	SV *	self = ST(0)
;
	int	response = (int)SvIV(ST(1))
;
	int	RETVAL;
	dXSTARG;
#line 512 "Fanotify.xs"
	int ret;
#line 837 "Fanotify.c"
#line 514 "Fanotify.xs"
	ret = _event_write_response(self, response);

	if (ret == -1) {
		XSRETURN_UNDEF;
	}

	RETVAL = ret;
#line 846 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_allow); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_allow)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	int	RETVAL;
	dXSTARG;
#line 531 "Fanotify.xs"
	int ret;
#line 866 "Fanotify.c"
#line 533 "Fanotify.xs"
	ret = _event_write_response(self, FAN_ALLOW);

	if (ret == -1) {
		XSRETURN_UNDEF;
	}

	RETVAL = ret;
#line 875 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_deny); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_deny)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	int	RETVAL;
	dXSTARG;
#line 550 "Fanotify.xs"
	int ret;
#line 895 "Fanotify.c"
#line 552 "Fanotify.xs"
	ret = _event_write_response(self, FAN_DENY);

	if (ret == -1) {
		XSRETURN_UNDEF;
	}

	RETVAL = ret;
#line 904 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event__dump); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event__dump)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
#line 569 "Fanotify.xs"
	struct fanotify_bundle *bundle;
	char buf[256];
#line 923 "Fanotify.c"
#line 572 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}
	printf("object dump --\n%s\n",
		event2str(bundle, buf, sizeof(buf)));
#line 930 "Fanotify.c"
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_Linux__Fanotify__Event__stringify); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event__stringify)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	SV *	RETVAL;
#line 585 "Fanotify.xs"
	struct fanotify_bundle *bundle;
	char buf[256];
	SV *s;
#line 950 "Fanotify.c"
#line 589 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}

	event2str(bundle, buf, sizeof(buf));

	s = newSVpvn_utf8(buf, strlen(buf), 1);
	// sv_2mortal(s); // XXX XXX Needed?

	RETVAL = s;
#line 962 "Fanotify.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_event_len); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_event_len)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	unsigned int	RETVAL;
	dXSTARG;
#line 609 "Fanotify.xs"
	struct fanotify_bundle *bundle;
#line 983 "Fanotify.c"
#line 611 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}
	RETVAL = bundle->metadata.event_len;
#line 989 "Fanotify.c"
	XSprePUSH; PUSHu((UV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_vers); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_vers)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	unsigned int	RETVAL;
	dXSTARG;
#line 625 "Fanotify.xs"
	struct fanotify_bundle *bundle;
#line 1009 "Fanotify.c"
#line 627 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}
	RETVAL = bundle->metadata.vers;
#line 1015 "Fanotify.c"
	XSprePUSH; PUSHu((UV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_reserved); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_reserved)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	unsigned int	RETVAL;
	dXSTARG;
#line 641 "Fanotify.xs"
	struct fanotify_bundle *bundle;
#line 1035 "Fanotify.c"
#line 643 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}
	RETVAL = bundle->metadata.reserved;
#line 1041 "Fanotify.c"
	XSprePUSH; PUSHu((UV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_metadata_len); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_metadata_len)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	unsigned int	RETVAL;
	dXSTARG;
#line 657 "Fanotify.xs"
	struct fanotify_bundle *bundle;
#line 1061 "Fanotify.c"
#line 659 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}
	RETVAL = bundle->metadata.metadata_len;
#line 1067 "Fanotify.c"
	XSprePUSH; PUSHu((UV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_mask); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_mask)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	uint64_t	RETVAL;
	dXSTARG;
#line 673 "Fanotify.xs"
	struct fanotify_bundle *bundle;
#line 1087 "Fanotify.c"
#line 675 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}
	RETVAL = bundle->metadata.mask;
#line 1093 "Fanotify.c"
	XSprePUSH; PUSHu((UV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_fd); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_fd)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	int	RETVAL;
	dXSTARG;
#line 689 "Fanotify.xs"
	struct fanotify_bundle *bundle;
#line 1113 "Fanotify.c"
#line 691 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}
	RETVAL = bundle->metadata.fd;
#line 1119 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_Linux__Fanotify__Event_pid); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_Linux__Fanotify__Event_pid)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "self");
    {
	SV *	self = ST(0)
;
	int	RETVAL;
	dXSTARG;
#line 705 "Fanotify.xs"
	struct fanotify_bundle *bundle;
#line 1139 "Fanotify.c"
#line 707 "Fanotify.xs"
	if (!(bundle = event2bundle(self))) {
		Perl_croak(aTHX_ "Invalid event object");
	}
	RETVAL = bundle->metadata.pid;
#line 1145 "Fanotify.c"
	XSprePUSH; PUSHi((IV)RETVAL);
    }
    XSRETURN(1);
}

#ifdef __cplusplus
extern "C"
#endif
XS_EXTERNAL(boot_Linux__Fanotify); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_Linux__Fanotify)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if (PERL_REVISION == 5 && PERL_VERSION < 9)
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        newXS_deffile("Linux::Fanotify::AUTOLOAD", XS_Linux__Fanotify_AUTOLOAD);
        (void)newXSproto_portable("Linux::Fanotify::fanotify_init", XS_Linux__Fanotify_fanotify_init, file, "$$");
        (void)newXSproto_portable("Linux::Fanotify::fanotify_mark", XS_Linux__Fanotify_fanotify_mark, file, "$$;$$$");
        (void)newXSproto_portable("Linux::Fanotify::fanotify_read", XS_Linux__Fanotify_fanotify_read, file, "$;$");
        (void)newXSproto_portable("Linux::Fanotify::fanotify_write", XS_Linux__Fanotify_fanotify_write, file, "$$");
        (void)newXSproto_portable("Linux::Fanotify::FanotifyGroup::getfd", XS_Linux__Fanotify__FanotifyGroup_getfd, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::FanotifyGroup::close", XS_Linux__Fanotify__FanotifyGroup_close, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::FanotifyGroup::DESTROY", XS_Linux__Fanotify__FanotifyGroup_DESTROY, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::close", XS_Linux__Fanotify__Event_close, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::DESTROY", XS_Linux__Fanotify__Event_DESTROY, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::needsResponse", XS_Linux__Fanotify__Event_needsResponse, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::_write", XS_Linux__Fanotify__Event__write, file, "$$");
        (void)newXSproto_portable("Linux::Fanotify::Event::allow", XS_Linux__Fanotify__Event_allow, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::deny", XS_Linux__Fanotify__Event_deny, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::_dump", XS_Linux__Fanotify__Event__dump, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::_stringify", XS_Linux__Fanotify__Event__stringify, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::event_len", XS_Linux__Fanotify__Event_event_len, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::vers", XS_Linux__Fanotify__Event_vers, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::reserved", XS_Linux__Fanotify__Event_reserved, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::metadata_len", XS_Linux__Fanotify__Event_metadata_len, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::mask", XS_Linux__Fanotify__Event_mask, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::fd", XS_Linux__Fanotify__Event_fd, file, "$");
        (void)newXSproto_portable("Linux::Fanotify::Event::pid", XS_Linux__Fanotify__Event_pid, file, "$");

    /* Initialisation Section */

#line 2 "./const-xs.inc"
  {
#if defined(dTHX) && !defined(PERL_NO_GET_CONTEXT)
    dTHX;
#endif
    HV *symbol_table = get_hv("Linux::Fanotify::", GV_ADD);

    static const struct iv_s values_for_iv[] =
      {
#ifdef FAN_ACCESS
        { "FAN_ACCESS", 10, FAN_ACCESS },
#endif
#ifdef FAN_MODIFY
        { "FAN_MODIFY", 10, FAN_MODIFY },
#endif
#ifdef FAN_CLOSE_WRITE
        { "FAN_CLOSE_WRITE", 15, FAN_CLOSE_WRITE },
#endif
#ifdef FAN_CLOSE_NOWRITE
        { "FAN_CLOSE_NOWRITE", 17, FAN_CLOSE_NOWRITE },
#endif
#ifdef FAN_OPEN
        { "FAN_OPEN", 8, FAN_OPEN },
#endif
#ifdef FAN_Q_OVERFLOW
        { "FAN_Q_OVERFLOW", 14, FAN_Q_OVERFLOW },
#endif
#ifdef FAN_OPEN_PERM
        { "FAN_OPEN_PERM", 13, FAN_OPEN_PERM },
#endif
#ifdef FAN_ACCESS_PERM
        { "FAN_ACCESS_PERM", 15, FAN_ACCESS_PERM },
#endif
#ifdef FAN_ONDIR
        { "FAN_ONDIR", 9, FAN_ONDIR },
#endif
#ifdef FAN_EVENT_ON_CHILD
        { "FAN_EVENT_ON_CHILD", 18, FAN_EVENT_ON_CHILD },
#endif
#ifdef FAN_CLOSE
        { "FAN_CLOSE", 9, FAN_CLOSE },
#endif
#ifdef FAN_CLOEXEC
        { "FAN_CLOEXEC", 11, FAN_CLOEXEC },
#endif
#ifdef FAN_NONBLOCK
        { "FAN_NONBLOCK", 12, FAN_NONBLOCK },
#endif
#ifdef FAN_CLASS_NOTIF
        { "FAN_CLASS_NOTIF", 15, FAN_CLASS_NOTIF },
#endif
#ifdef FAN_CLASS_CONTENT
        { "FAN_CLASS_CONTENT", 17, FAN_CLASS_CONTENT },
#endif
#ifdef FAN_CLASS_PRE_CONTENT
        { "FAN_CLASS_PRE_CONTENT", 21, FAN_CLASS_PRE_CONTENT },
#endif
#ifdef FAN_ALL_CLASS_BITS
        { "FAN_ALL_CLASS_BITS", 18, FAN_ALL_CLASS_BITS },
#endif
#ifdef FAN_UNLIMITED_QUEUE
        { "FAN_UNLIMITED_QUEUE", 19, FAN_UNLIMITED_QUEUE },
#endif
#ifdef FAN_UNLIMITED_MARKS
        { "FAN_UNLIMITED_MARKS", 19, FAN_UNLIMITED_MARKS },
#endif
#ifdef FAN_ALL_INIT_FLAGS
        { "FAN_ALL_INIT_FLAGS", 18, FAN_ALL_INIT_FLAGS },
#endif
#ifdef FAN_MARK_ADD
        { "FAN_MARK_ADD", 12, FAN_MARK_ADD },
#endif
#ifdef FAN_MARK_REMOVE
        { "FAN_MARK_REMOVE", 15, FAN_MARK_REMOVE },
#endif
#ifdef FAN_MARK_DONT_FOLLOW
        { "FAN_MARK_DONT_FOLLOW", 20, FAN_MARK_DONT_FOLLOW },
#endif
#ifdef FAN_MARK_ONLYDIR
        { "FAN_MARK_ONLYDIR", 16, FAN_MARK_ONLYDIR },
#endif
#ifdef FAN_MARK_MOUNT
        { "FAN_MARK_MOUNT", 14, FAN_MARK_MOUNT },
#endif
#ifdef FAN_MARK_IGNORED_MASK
        { "FAN_MARK_IGNORED_MASK", 21, FAN_MARK_IGNORED_MASK },
#endif
#ifdef FAN_MARK_IGNORED_SURV_MODIFY
        { "FAN_MARK_IGNORED_SURV_MODIFY", 28, FAN_MARK_IGNORED_SURV_MODIFY },
#endif
#ifdef FAN_MARK_FLUSH
        { "FAN_MARK_FLUSH", 14, FAN_MARK_FLUSH },
#endif
#ifdef FAN_ALL_MARK_FLAGS
        { "FAN_ALL_MARK_FLAGS", 18, FAN_ALL_MARK_FLAGS },
#endif
#ifdef FAN_ALL_EVENTS
        { "FAN_ALL_EVENTS", 14, FAN_ALL_EVENTS },
#endif
#ifdef FAN_ALL_PERM_EVENTS
        { "FAN_ALL_PERM_EVENTS", 19, FAN_ALL_PERM_EVENTS },
#endif
#ifdef FAN_ALL_OUTGOING_EVENTS
        { "FAN_ALL_OUTGOING_EVENTS", 23, FAN_ALL_OUTGOING_EVENTS },
#endif
#ifdef FANOTIFY_METADATA_VERSION
        { "FANOTIFY_METADATA_VERSION", 25, FANOTIFY_METADATA_VERSION },
#endif
#ifdef FAN_ALLOW
        { "FAN_ALLOW", 9, FAN_ALLOW },
#endif
#ifdef FAN_DENY
        { "FAN_DENY", 8, FAN_DENY },
#endif
#ifdef FAN_NOFD
        { "FAN_NOFD", 8, FAN_NOFD },
#endif
        { NULL, 0, 0 } };
	const struct iv_s *value_for_iv = values_for_iv;
        while (value_for_iv->name) {
	    constant_add_symbol(aTHX_  symbol_table, value_for_iv->name,
				value_for_iv->namelen, newSViv(value_for_iv->value));
            ++value_for_iv;
	}
	if (C_ARRAY_LENGTH(values_for_notfound) > 1) {
#ifndef SYMBIAN
	    HV *const constant_missing = get_missing_hash(aTHX);
#endif
	    const struct notfound_s *value_for_notfound = values_for_notfound;
	    do {

		/* Need to add prototypes, else parsing will vary by platform.  */
		HE *he = (HE*) hv_common_key_len(symbol_table,
						 value_for_notfound->name,
						 value_for_notfound->namelen,
						 HV_FETCH_LVALUE, NULL, 0);
		SV *sv;
#ifndef SYMBIAN
		HEK *hek;
#endif
		if (!he) {
		    croak("Couldn't add key '%s' to %%Linux::Fanotify::",
			  value_for_notfound->name);
		}
		sv = HeVAL(he);
		if (!SvOK(sv) && SvTYPE(sv) != SVt_PVGV) {
		    /* Nothing was here before, so mark a prototype of ""  */
		    sv_setpvn(sv, "", 0);
		} else if (SvPOK(sv) && SvCUR(sv) == 0) {
		    /* There is already a prototype of "" - do nothing  */
		} else {
		    /* Someone has been here before us - have to make a real
		       typeglob.  */
		    /* It turns out to be incredibly hard to deal with all the
		       corner cases of sub foo (); and reporting errors correctly,
		       so lets cheat a bit.  Start with a constant subroutine  */
		    CV *cv = newCONSTSUB(symbol_table,
					 value_for_notfound->name,
					 &PL_sv_yes);
		    /* and then turn it into a non constant declaration only.  */
		    SvREFCNT_dec(CvXSUBANY(cv).any_ptr);
		    CvCONST_off(cv);
		    CvXSUB(cv) = NULL;
		    CvXSUBANY(cv).any_ptr = NULL;
		}
#ifndef SYMBIAN
		hek = HeKEY_hek(he);
		if (!hv_common(constant_missing, NULL, HEK_KEY(hek),
 			       HEK_LEN(hek), HEK_FLAGS(hek), HV_FETCH_ISSTORE,
			       &PL_sv_yes, HEK_HASH(hek)))
		    croak("Couldn't add key '%s' to missing_hash",
			  value_for_notfound->name);
#endif
	    } while ((++value_for_notfound)->name);
	}
    /* As we've been creating subroutines, we better invalidate any cached
       methods  */
    mro_method_changed_in(symbol_table);
  }

#line 1385 "Fanotify.c"

    /* End of Initialisation Section */

#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

