#define PERL_constant_NOTFOUND	1
#define PERL_constant_NOTDEF	2
#define PERL_constant_ISIV	3
#define PERL_constant_ISNO	4
#define PERL_constant_ISNV	5
#define PERL_constant_ISPV	6
#define PERL_constant_ISPVN	7
#define PERL_constant_ISSV	8
#define PERL_constant_ISUNDEF	9
#define PERL_constant_ISUV	10
#define PERL_constant_ISYES	11

#ifndef NVTYPE
typedef double NV; /* 5.6 and later define NVTYPE, and typedef NV to it.  */
#endif
#ifndef aTHX_
#define aTHX_ /* 5.6 or later define this for threading support.  */
#endif
#ifndef pTHX_
#define pTHX_ /* 5.6 or later define this for threading support.  */
#endif

/* This allows slightly more efficient code on !USE_ITHREADS: */
#ifdef USE_ITHREADS
#  define COP_FILE(c)	CopFILE(c)
#  define COP_FILE_F	"s"
#else
#  define COP_FILE(c)	CopFILESV(c)
#  define COP_FILE_F	SVf
#endif

static void
constant_add_symbol(pTHX_  HV *hash, const char *name, I32 namelen, SV *value) {
    HE *he = (HE*) hv_common_key_len(hash, name, namelen, HV_FETCH_LVALUE, NULL,
				     0);
    SV *sv;

    if (!he) {
        croak("Couldn't add key '%s' to %%Linux::Fanotify::",
		   name);
    }
    sv = HeVAL(he);
    if (SvOK(sv) || SvTYPE(sv) == SVt_PVGV) {
	/* Someone has been here before us - have to make a real sub.  */
	newCONSTSUB(hash, name, value);
    } else {
	SvUPGRADE(sv, SVt_RV);
	SvRV_set(sv, value);
	SvROK_on(sv);
	SvREADONLY_on(value);
    }
}



#ifndef SYMBIAN

/* Store a hash of all symbols missing from the package. To avoid trampling on
   the package namespace (uninvited) put each package's hash in our namespace.
   To avoid creating lots of typeblogs and symbol tables for sub-packages, put
   each package's hash into one hash in our namespace.  */

static HV *
get_missing_hash(pTHX) {
    HV *const parent
	= get_hv("ExtUtils::Constant::ProxySubs::Missing", GVf_MULTI);
    /* We could make a hash of hashes directly, but this would confuse anything
	at Perl space that looks at us, and as we're visible in Perl space,
	best to play nice. */
    SV *const *const ref
	= hv_fetch(parent, "Linux::Fanotify", 15, TRUE);
    HV *new_hv;

    if (!ref)
	return NULL;

    if (SvROK(*ref))
	return (HV*) SvRV(*ref);

    new_hv = newHV();
    SvUPGRADE(*ref, SVt_RV);
    SvRV_set(*ref, (SV *)new_hv);
    SvROK_on(*ref);
    return new_hv;
}

#endif

struct notfound_s {const char *name; I32 namelen;} ;

    static const struct notfound_s values_for_notfound[] =
      {
#ifndef FAN_ACCESS
        { "FAN_ACCESS", 10 },
#endif
#ifndef FAN_MODIFY
        { "FAN_MODIFY", 10 },
#endif
#ifndef FAN_CLOSE_WRITE
        { "FAN_CLOSE_WRITE", 15 },
#endif
#ifndef FAN_CLOSE_NOWRITE
        { "FAN_CLOSE_NOWRITE", 17 },
#endif
#ifndef FAN_OPEN
        { "FAN_OPEN", 8 },
#endif
#ifndef FAN_Q_OVERFLOW
        { "FAN_Q_OVERFLOW", 14 },
#endif
#ifndef FAN_OPEN_PERM
        { "FAN_OPEN_PERM", 13 },
#endif
#ifndef FAN_ACCESS_PERM
        { "FAN_ACCESS_PERM", 15 },
#endif
#ifndef FAN_ONDIR
        { "FAN_ONDIR", 9 },
#endif
#ifndef FAN_EVENT_ON_CHILD
        { "FAN_EVENT_ON_CHILD", 18 },
#endif
#ifndef FAN_CLOSE
        { "FAN_CLOSE", 9 },
#endif
#ifndef FAN_CLOEXEC
        { "FAN_CLOEXEC", 11 },
#endif
#ifndef FAN_NONBLOCK
        { "FAN_NONBLOCK", 12 },
#endif
#ifndef FAN_CLASS_NOTIF
        { "FAN_CLASS_NOTIF", 15 },
#endif
#ifndef FAN_CLASS_CONTENT
        { "FAN_CLASS_CONTENT", 17 },
#endif
#ifndef FAN_CLASS_PRE_CONTENT
        { "FAN_CLASS_PRE_CONTENT", 21 },
#endif
#ifndef FAN_ALL_CLASS_BITS
        { "FAN_ALL_CLASS_BITS", 18 },
#endif
#ifndef FAN_UNLIMITED_QUEUE
        { "FAN_UNLIMITED_QUEUE", 19 },
#endif
#ifndef FAN_UNLIMITED_MARKS
        { "FAN_UNLIMITED_MARKS", 19 },
#endif
#ifndef FAN_ALL_INIT_FLAGS
        { "FAN_ALL_INIT_FLAGS", 18 },
#endif
#ifndef FAN_MARK_ADD
        { "FAN_MARK_ADD", 12 },
#endif
#ifndef FAN_MARK_REMOVE
        { "FAN_MARK_REMOVE", 15 },
#endif
#ifndef FAN_MARK_DONT_FOLLOW
        { "FAN_MARK_DONT_FOLLOW", 20 },
#endif
#ifndef FAN_MARK_ONLYDIR
        { "FAN_MARK_ONLYDIR", 16 },
#endif
#ifndef FAN_MARK_MOUNT
        { "FAN_MARK_MOUNT", 14 },
#endif
#ifndef FAN_MARK_IGNORED_MASK
        { "FAN_MARK_IGNORED_MASK", 21 },
#endif
#ifndef FAN_MARK_IGNORED_SURV_MODIFY
        { "FAN_MARK_IGNORED_SURV_MODIFY", 28 },
#endif
#ifndef FAN_MARK_FLUSH
        { "FAN_MARK_FLUSH", 14 },
#endif
#ifndef FAN_ALL_MARK_FLAGS
        { "FAN_ALL_MARK_FLAGS", 18 },
#endif
#ifndef FAN_ALL_EVENTS
        { "FAN_ALL_EVENTS", 14 },
#endif
#ifndef FAN_ALL_PERM_EVENTS
        { "FAN_ALL_PERM_EVENTS", 19 },
#endif
#ifndef FAN_ALL_OUTGOING_EVENTS
        { "FAN_ALL_OUTGOING_EVENTS", 23 },
#endif
#ifndef FANOTIFY_METADATA_VERSION
        { "FANOTIFY_METADATA_VERSION", 25 },
#endif
#ifndef FAN_ALLOW
        { "FAN_ALLOW", 9 },
#endif
#ifndef FAN_DENY
        { "FAN_DENY", 8 },
#endif
#ifndef FAN_NOFD
        { "FAN_NOFD", 8 },
#endif
        { NULL, 0 } };
struct iv_s {const char *name; I32 namelen; IV value;};
